/********************************************************************
* SPEECH1 - collection of functions for speech processing
* 
* Author:  		Jan Cernocky
* Created: 		16/10/97
* Last modification:	17/3/98 - median filter
*                       1/4/98  - path in DTW alignment
*                       Fri Sep  8 13:47:16 CEST 2000 - linear warp
*                       Sun Oct 13 23:42:20 CEST 2002 - many functions on lpc
                           tested in SPEECH1_TEST - looks like ok. 
*
* Description: 
*   Various functions for speech processing. They should be as general as
*   possible, without global parameters. The functions should be well 
*   commented. Later will possibly include functions from UNIANAL
*   directory.
* Uses:
*   common.c for standard tasks (alloc, free, etc)
* 
* Original file speech.c was too specialized.
*********************************************************************/

#include <stdio.h>
#include <math.h>
#include "common.h"
#include "speech1.h"

/*#define DEBUG_MEDIAN*/
/*#define DEBUG_DTW*/

  /*********************************************************************
   * FUNCTION: SqEuclidDistance
   * PURPOSE: squqred Euclidian Distance
   * INPUTS: 
   *    pfA[0..swLen-1]      1st vector
   *    pfB[0..swLen-1]      2nd vector
   *    swLen                their length
   * OUTPUTS: 
   *    none
   * RETURN: 
   *    squarred distance (float)   
   * DESCRIPTION:
   * KEYWORDS:
   ***********************************************************************/
float SqEuclidDistance (float *pfA, float *pfB, short swLen)
{
   short i;
   float *f,*g,aux, sum;
   
   sum=0.0; f=pfA; g=pfB;
   for (i=0; i< swLen; i++)  {
      aux = (*f++) - (*g++);
      sum += aux * aux;
   }
   return (sum);
}

  /*********************************************************************
   * FUNCTION: DTWDistance
   * PURPOSE: Dynamic Time Warping Distance
   * INPUTS: 
   *    ppfR[0..swLenR-1, 0..swSize-1] - 	reference param matrix
   *    ppfT[0..swLenT-1, 0..swSize-1] -	test param matrix
   *    swLenR			       -   	ref. length
   *    swLenT			       -	test length
   *    swSize			       -	par. vec. size
   *    pswPath - vector allocated for path, if NULL, the path won't be
   *              written
   * OUTPUTS: 
   *    pswPath[0..swLen..... - DTW alignment path 
   * RETURN: 
   *    DTW (float)   
   * DESCRIPTION:
   *    based on Matlab function of Pascale Jardin & Gen. Baudoin
   * KEYWORDS:
   ***********************************************************************/
float DTWDistance (float **ppfR, float **ppfT, short swLenR, short swLenT, 
		   short swSize, short *pswPath)
{
   float **g,d,g1,g2,g3;
   short i,j,**trace, ir,it;
   
   g=Alloc_2_Float (swLenT+1, swLenR+1);     /* cumul dist */
   trace=Alloc_2_Short (swLenT+1, swLenR+1); /* for path backtrace */
   /* fill the 1st row and col. with big values, except for [0,0] */
   for (i=1; i<=swLenR; i++) g[0][i]=BIG_FLOAT_VALUE;
   for (i=1; i<=swLenT; i++) g[i][0]=BIG_FLOAT_VALUE;
   g[0][0]=0.0;
   /* now go and make the cumul dist */
   for (i=1; i<=swLenT; i++) {
      for (j=1; j<=swLenR; j++)  {
	 d=SqEuclidDistance(ppfT[i-1], ppfR[j-1], swSize);
	 g2=g[i-1][j] + d;
	 g3=g[i][j-1] + d;
	 g1=g[i-1][j-1] + d + d; /* faster than 2*d */
	 if (g1 < g2 && g1 < g3) {
	   g[i][j] = g1;
	   trace[i][j] = 1;
	 }
	 else
	   if (g2 < g3) {
	      g[i][j] = g2;
	      trace[i][j] = 2;
	   }
	   else {
	      g[i][j] = g3;
	      trace[i][j] = 3;
	   }
      }
   }
#ifdef DEBUG_DTW
   for(i=0; i<=swLenT; i++) pfv ("", g[i], swLenR+1); pnl();
   for(i=0; i<=swLenT; i++) pswv ("", trace[i], swLenR+1); pnl();
#endif      
   
   /* --------------------------------------------- */ 
   /* if the path requested, going to backtrace it  */
   /* --------------------------------------------- */

   if (pswPath != NULL ) {
     ir=swLenR; it=swLenT;    /* init for backtrace */
     pswPath[ir-1] = it-1;    /* last vector always aligned with last */
 
     while (it + ir > 2) {   /* last point: it=1 and ir=1 */
       /* --- depending on the trace, modif. the pointers --- */
       switch (trace[it][ir]) {
       case 1: it--; ir--; break;
       case 2: it--;       break;
       case 3:       ir--; break;
       }
       /* --- and record to Path --- */
       pswPath[ir-1] = it-1;
     } /* of while */
     
     /* --- the last vector should be always aligned with the last, so
            impose it --- */
     pswPath[swLenR-1] = swLenT-1;

   } /* of path creation */

   d=g[swLenT][swLenR] / (float)(swLenT + swLenR);
   Free_2_Float (g, swLenT+1);
   Free_2_Short (trace, swLenT+1); 
   return (d);
}

  /*********************************************************************
   * FUNCTION: ScalarQuant
   * PURPOSE: Scalar Quantization using codebook
   * INPUTS: 
   *    pfCB[0..swL-1]		SQ codebook
   *    swL			its length
   *    fInput			value to be quantized
   * OUTPUTS: 
   *    pswIndex		pointer to index of quantized value in CB
   * RETURN: 
   *    fOutput (float) quantized input
   * DESCRIPTION:
   *    simple implementation, computes distance from all members, 
   *    chooses the min.
   * KEYWORDS:
   ***********************************************************************/
float ScalarQuant (float *pfCB, short swL, float fInput, short *pswIndex)
{
   short i, imin;
   float d, dmin;
   
   dmin = 1e36;
   for (i=0; i<swL; i++)  {
      d = fInput - pfCB [i];
      d = (d>0.0 ? d : -d );
      if (d < dmin)  {
	 dmin = d;
	 imin = i;
      }
   }
   *pswIndex = imin;
   return (pfCB[imin]);
}
   
 /*********************************************************************
   * FUNCTION: MedianFilter
   * PURPOSE: Median filtering of vector of values.
   * INPUTS: 
   *    pfIn[0..L_L-1]		input signal
   *    L_L			its length
   *    swM			length of median filter. Must be odd.
   * OUTPUTS: 
   *    pfOut[0..L_L-1]		output signal (must be already allocated)
   * RETURN: 
   *    none
   * DESCRIPTION:
   *    Do not the change the length within a program, there's an allocation
   *	in 1st pass of this function ! The filter is not causal ! 
   *    No optimization.
   * KEYWORDS:
   ***********************************************************************/
void MedianFilter (float *pfIn, long L_L, short swM, float *pfOut)
{
   static short swFirstPass=1;
   static float *pfBuffer;
   float *pfAux1, *pfAux2;
   long L_From, L_To, il, jl;
   short swRealLength, swHalfM, js;
   
   /*** if 1st pass, alloc for buffer ***/
   if (swFirstPass == 1)  {
      pfBuffer = Alloc_1_Float (swM);
      swFirstPass = 0;
   }
   /*** get half length of filter ***/
   swHalfM=swM / 2;
   
   for (il = 0; il<L_L; il++)  {
      /* determine, where begin and where end */
      L_From = (il-swHalfM >= 0) ? il-swHalfM : 0;
      L_To   = (il+swHalfM < L_L) ? il+swHalfM : L_L-1;
      /* get the real length of filter and copy to buffer */
      swRealLength = L_To - L_From + 1;
      pfAux1 = &pfIn [L_From]; pfAux2 = pfBuffer;
      for (js=0; js<swRealLength; js++)
      	*(pfAux2++) = *(pfAux1++);
      /* sort it */
#ifdef DEBUG_MEDIAN
      pfv("notsort",pfBuffer,swRealLength );
#endif        
      Sort (pfBuffer, NULL, swRealLength);
      /* get the median value */
      pfOut[il] = pfBuffer [swRealLength / 2];
#ifdef DEBUG_MEDIAN
      pfv("sorted",pfBuffer,swRealLength );
      pLs ("L_From",L_From); pLs ("L_To",L_To); psws ("swRealLength",swRealLength);
      pfs ("median", pfBuffer [swRealLength / 2]); pnl();
#endif      

   }
}
      
 /*********************************************************************
   * FUNCTION: MeanFilter
   * PURPOSE: Mean filtering of vector of values.
   * INPUTS: 
   *    pfIn[0..L_L-1]		input signal
   *    L_L			its length
   *    swM			length of mean filter. Must be odd.
   * OUTPUTS: 
   *    pfOut[0..L_L-1]		output signal (must be already allocated)
   * RETURN: 
   *    none
   * DESCRIPTION:
   *    for sample, looks at M-1 neighbours and computes the mean. It is 
   *    good for some smoothing. Not optimized.
   * KEYWORDS:
   ***********************************************************************/
void MeanFilter (float *pfIn, long L_L, short swM, float *pfOut)
{
   float *pfBuffer, fMean;
   long L_From, L_To, il, jl;
   short swRealLength, swHalfM, js;
   
   /*** get half length of filter ***/
   swHalfM=swM / 2;
   
   for (il = 0; il<L_L; il++)  {
      /* determine, where begin and where end */
      L_From = (il-swHalfM >= 0) ? il-swHalfM : 0;
      L_To   = (il+swHalfM < L_L) ? il+swHalfM : L_L-1;
      /* get the real length of filter and pointer to 1st sample  */
      swRealLength = L_To - L_From + 1;
      pfBuffer = &pfIn [L_From]; 
      /* compute the mean and write it down */
      fMean = Mean (pfBuffer, swRealLength);
      pfOut[il] = fMean;
   }
}
      
 /*********************************************************************
   * FUNCTION: LinearWarping
   * PURPOSE: produces sequence of time indices for linear time warp
   * INPUTS: 
   *    swLenRef		reference Length
   *    swLenTest		test length
   *    pswPath			allocated with length swLenRef
   * OUTPUTS: 
   *    pswPath		      filled with the path going from 0 to swLenTest-1
   * RETURN: 
   *    none
   * DESCRIPTION:
   * KEYWORDS:
   ***********************************************************************/
void LinearWarping (short swLenRef, short swLenTest, short *pswPath) 
{
  float ratio, refval;
  short is;
  
  ratio = (float)(swLenTest - 1) / (float)(swLenRef - 1);
  for (is = 0; is<swLenRef; is++){
    pswPath[is] = (short)(rint((float)is * ratio));
  }
}

   /*********************************************************************
   * FUNCTION: CenterVector
   * PURPOSE: subtracts the mean of a vector
   * INPUTS: 
   *	pfV[0:L_LengthV] - vector
   *    L_LengthV - its length
   * OUTPUTS: 
   *    pfV[0:L_LengthV] - centered vector
   *
   * RETURN: 
   *	none   
   *
   * DESCRIPTION:
   * KEYWORDS:
   ***********************************************************************/
void CenterVector (float *pfV, long L_LengthV)
{
   float fSum, fMean;
   long li;
   
   fSum=0;
   for (li=0; li < L_LengthV; li++) 
      fSum += pfV [li];
   fMean = fSum / (float)L_LengthV;
   
   if (fMean != 0.0)
      for (li=0; li < L_LengthV; li++)
	 pfV[li] -= fMean;
}
   
   /*********************************************************************
   * FUNCTION: EnergyNoWindow
   * PURPOSE: energy of signal (related to 1 sample)
   * INPUTS: 
   *	pfSFrame[0:Lseg-1] - speech frame
   *    swLseg - segment length
   *
   * OUTPUTS: 
   *	none
   *
   * RETURN:                      N-1
   *	fSE - signal energy = 1/N sum x(n)^2   
   *                              n=0
   * DESCRIPTION:
   * KEYWORDS:
   ***********************************************************************/

float EnergyNoWindow (float *pfSFrame, short swLseg)
{
   short i;
   float sum;

   sum = 0.0;
   for (i=0; i<swLseg; i++)
	 sum += pfSFrame[i] * pfSFrame[i];
   
   /* --- get the total energy of frame divided by the nb. of samples --- */
   
   return(sum / (float)swLseg);
}

   /********************************************************************
   * FUNCTION: Autocorrelation
   * PURPOSE: calculates the autocor. coeffs R
   * INPUTS:  
   *	pfWFrame[0:Lseg-1] - speech frame with window
   *    swLseg - segment length
   *	swLpcOrder - order of Lpc analysis
   *
   * OUTPUT: 
   *	pfR[0:swLpcOrder] - autocorrelation coeffs.
   *
   * RETURN: none
   * DESCRIPTION:
   * KEYWORDS:
   * AUTHOR: Pascalle Jardin, modified by Jan C.
   *********************************************************************/

void Autocorrelation(float *pfWFrame, float *pfR, short swLseg, short swLpcOrder)
{
   float corre ;
   int i,k;
   float *dat;
   
   corre=0.0;
   dat = pfWFrame;
   
   for(i=0; i<swLseg; i++, dat++)
      corre += (*dat) * *dat;
   pfR[0] = corre;
   
   for(k=1; k<=swLpcOrder; k++){
      dat = pfWFrame; 
      for(i=0,corre =0.0,dat = pfWFrame; i<swLseg-k; i++,dat++)
	 corre += *dat * *(dat+k) ;
      pfR[k] = corre;
   }
}


   /*********************************************************************
   * FUNCTION: HammingFrame
   * PURPOSE: window speech by hamming window.
   * INPUTS: 
   *	pfSFrame[0:Lseg-1] - speech frame
   *    swLseg - segment length
   * OUTPUTS: 
   *    pfWSFrame[0:Lseg-1] - windowed frame
   *
   * RETURN: 
   *	none   
   *
   * DESCRIPTION:
   * KEYWORDS:
   ***********************************************************************/

void HammingFrame  (float *pfSFrame, float *pfWSFrame, short swLseg)
{
   static float *pfWindow;
   static short swFirstPass=1;
   float *pf1, *pf2, *pfw;
   short swI;
   
   if (swFirstPass) {   /* must allocate meme and init. the window */
      pfWindow = Alloc_1_Float (swLseg);
      
      for (swI=0; swI < swLseg; swI++)
	pfWindow[swI] = 0.54-0.46*cos(2*PI*(float)swI / (float)(swLseg-1));
      swFirstPass = 0;
   }
   
   pf1 = pfSFrame; pf2 = pfWSFrame; pfw = pfWindow;
   for (swI=0; swI < swLseg; swI++)
      *(pf2++) = *(pf1++) * *(pfw++);
}

   /********************************************************************
   * FUNCTION: Leroux
   * PURPOSE: calculates the refl. coeffs from autocor. coeffs
   * INPUTS:  
   *	pfR[0:swLpcOrder] - autocorrelation coeffs.
   *	swLpcOrder - order of Lpc analysis
   *
   * OUTPUT: 
   *	pfK[0:swLpcOrder-1] - reflection coeffs. if k-th coeff. is bad,
   *                           k and the rest are zeroed
   *
   * RETURN: 
   *	swBadFr - 0 if o.k., 1 if bad (unstable) coeff.
   * DESCRIPTION:
   * KEYWORDS:
   * AUTHOR: Pascalle Jardin, modified by Jan C.
   *********************************************************************/

short Leroux (float *pfR, float *pfK, short swLpcOrder)
{
   short i,m,j;
   static float *en, *ep;
   float epp;
   static short swFirstPass = 1;
   
   if (swFirstPass) {
      en = Alloc_1_Float (swLpcOrder + 1);
      ep = Alloc_1_Float (swLpcOrder + 1);
      swFirstPass=0;
   }
   
   for (i=1;i<= swLpcOrder;i++){
      en[i]= pfR[i-1] ;
      ep[i]= pfR[i];
   }
   
   for(m=0;m<swLpcOrder;m++){
      if(fabs(ep[m+1]) >= en[1]){
/*	 printf("incident leroux ordre %d\n",m); */
/*	 printf("abs(ep[%d])=%f > en[1]= %f\n",m+1,fabs(ep[m+1]),en[1]); */
/* === bad coefficient - zero all ! */	 
	 for(j=0;j<swLpcOrder;j++)
	    pfK[j] = 0.0;
	 return (1);			/* bad frame */
      }
      else{
	 pfK[m] = -ep[m+1]/en[1];
	 for(i=m+1;i<=swLpcOrder;i++){
	    epp = ep[i];
	    ep[i] += (en[i-m] * pfK[m]);
	    en[i-m] += (epp * pfK[m]);
	 }
      }
   }
   return(0);	/* o.k. */
}

   /********************************************************************
   * FUNCTION: ResidualEnergy
   * PURPOSE: calculates the energ of residual
   * INPUTS:  
   *	fSigEnergy - energy of signal frame (without windowing)
   *	pfK[0:swLpcOrder-1] - reflection coeffs.
   *	swLpcOrder - order of Lpc analysis
   *
   * OUTPUT: 
   *	none
   *
   * RETURN: 
   *	fResEnergy - energy of the residual
   *
   * DESCRIPTION:
   * KEYWORDS:
   *********************************************************************/

float ResidualEnergy (float fSE, float *pfK, short swLpcOrder)
{
   float fProduct, fResEnergy;
   short i;
   
   fProduct = 1.0;
   for (i=0; i< swLpcOrder; i++)
      fProduct *= 1 - pfK[i] * pfK[i];
   fResEnergy = fSE * fProduct;
   
   return (fResEnergy);
}

   /********************************************************************
   * FUNCTION: RflToLpc
   * PURPOSE: reflection -> lpc coefficients
   * INPUTS:  
   *	pfR[0:swLpcOrder-1] - refl. coeffs.
   *	swLpcOrder - order of Lpc analysis
   *
   * OUTPUT: 
   *	pfA[0:swLpcOrder-1] - lpc coeffs. a1...aP
   *
   * RETURN: 
   *	none
   *
   * DESCRIPTION:
   * KEYWORDS:
   *********************************************************************/

void RflToLpc (float *pfR, float *pfA, short swLpcOrder)
{
   short m,i,j;
   static float *B, *Bm;
   static short swFirstPass=1;
   
   if (swFirstPass) {
      B = Alloc_1_Float (swLpcOrder + 1);
      Bm = Alloc_1_Float (swLpcOrder + 1);
      swFirstPass = 0 ;
   }	 

   Bm[0]=1;
   for (m=1;m < swLpcOrder+1; m++)  
   {
      B[0]=Bm[0];
      for(i=1;i<m;i++) B[i]=Bm[i]+pfR[m-1]*Bm[m-i];
      B[m]=pfR[m-1];
      for(i=0;i<=m;i++) Bm[i]=B[i];
   }
   for(m=0; m < swLpcOrder; m++)
      pfA[m]=B[m+1];		/* output without a0=1 */
}

   /********************************************************************
   * FUNCTION: LpcToCps
   * PURPOSE: lpc -> lpcc coefficients
   * INPUTS:  
   *    pfA[0:swLpcOrder-1] - lpc coeffs. a1...aP
   *    swLpcOrder - order of Lpc analysis
   *    swCpsNumber - number of lpcc coeffs to compute
   *
   * OUTPUT: 
   *    pfC[0:swCpsNumber-1] - lpcc coeffs. c1...cQ
   *
   * RETURN: 
   *    none
   *
   * DESCRIPTION:
   * KEYWORDS:
   *********************************************************************/

void LpcToCps (float *pfA, float *pfC, short swLpcOrder, short swCpsNumber)
{
   short i,j,iindex,jindex;
   float sum;
   static float *help;
   static short swFirstPass=1;
   
   if (swFirstPass) {
      help=Alloc_1_Float (swCpsNumber);
      swFirstPass = 0;
   }
   
   for (i=0; i < (swCpsNumber < swLpcOrder ? swCpsNumber : swLpcOrder); i++)
      help[i]=pfA[i];
   for (j=i; j< swCpsNumber; j++)   /* fill the rest (if any) with zeros */
      help[j]=0.0;

   for (i=1, iindex=0; i <= swCpsNumber; i++, iindex++) { /* i,j - theoretical
                                                             indices, xindex -
                                                             addressing */
      sum=0.0;
      for(j=1, jindex=0; j <= (i-1); j++, jindex++)
         sum += ( (float)j / (float)i ) * help[i-j-1]*pfC[jindex];
      pfC[iindex] = - help[iindex] - sum;
   }
}
